'''
Runners for handling spiro-managed saltenvs
'''
from __future__ import absolute_import, print_function, unicode_literals
import logging
from salt.ext import six


log = logging.getLogger(__name__)


def query_highstate(saltenv):
    """
    Return the list of minions that have tops in the given saltenv.

    If the given saltenv has changed, these minions might be effected.

    Note that this is generally useful and does not require the use of spirofs.

    Note: This assumes you do not have any Jinja in your top files that mutates your systems.
    """
    tops = __salt__['salt.execute']('*', 'state.show_top', kwarg={'concurrent': True})
    return [
        mid
        for mid, envs in tops.items()
        if envs and saltenv in envs
    ]


def issue_token(projects, lifespan=None):
    """
    Issue a deployment token for the given projects, optionally for the given time (in seconds)
    """
    if isinstance(projects, six.string_types):
        projects = projects.split(',')

    if lifespan is not None:
        lifespan = int(lifespan)

    return __utils__['spiro_auth.generate_token'](projects=projects, lifespan=lifespan)


def issue_root_token(lifespan=None):
    """
    Issue a deployment token for all projects, optionally for the given time (in seconds)
    """
    if lifespan is not None:
        lifespan = int(lifespan)

    return __utils__['spiro_auth.generate_token'](lifespan=lifespan)


def revoke_token(token):
    """
    Revokes the given token, making it invalid and disallowing it from accessing
    anything.
    """
    log.warning("TODO: Implement revoke_token")


def check_token(token, project=None):
    """
    Checks if the given token is valid for the given project. If not project is
    given, only check that the token is valid.
    """
    return __utils__['spiro_auth.check_token'](token, project)


def inspect_token(token):
    """
    Gets the information embedded in the given token in a human-readable format.
    """
    return __utils__['spiro_auth.inspect_token'](token)




# -*- coding: utf-8 -*-
'''
Spiro CI receiving server.

Takes the following configuration arguments:
* port: Port to listen on, defaults to 4510
* ssl_key: The private key for TLS
* ssl_crt: The public certificate for TLS

If ssl_key/ssl_crt are not given, plain HTTP without TLS is used. If one is
given, that is an error.

Creates an HTTP server at the given port. It has one path, /, which accepts two
methods:

* GET: Checks the given token is valid, returning either a 401 Unauthorized or
  200 OK
* POST: Submits a deployment. Take form-like data (multipart/form-data):
  * project: The name of the project
  * deployment: The name of the deployment for the given project
  * highstate: If given, highstates affected minions (ie those with topfile
    entries for the given deployment)
  * bundle: Uploaded file: an uncompressed tar containing the files that make up
    the new deployment

The POST returns a text/event-stream, with the caveat that it is not resumable.
If the connection is left open, deployment updates are sent to the client where
the event is a code representing the stage and the data is a JSON. All messages
include a msg property, which is the human-readable status.

Deployment stages are:
* deploy-bundle: The uploaded bundle is being extracted and any generation that
  happens is being done, args: msg
* highstate-calc: Calculating minions to highstate, args: msg
* highstate-start: We're begining to highstate minions, args: msg, minions
* highstate-update: A minion has finished its highstate, args: msg, minion,
  result (big blob of highstate data)
* highstate-finish: The highstate has completed, args: msg

If there's an error, an exception message is sent with the args msg.

Authentication is HTTP Bearer auth, eg the client should be configured with a
token and send a header in the form of "Authorization: Bearer <token>". If no
token is given or is invalid (authentication fails), a 401 Unauthorized is
returned. If authorization fails (the given token is not allowed to deploy to
the given project), a 403 Forbidden is returned.
'''
from __future__ import absolute_import, print_function, unicode_literals
import tarfile
import logging
import tornado.concurrent
import tornado.ioloop
import tornado.iostream
import tornado.web
import tornado.gen
import tornado.queues
import tornado.httpserver
import json
import ssl
import os
import traceback
from salt.ext import six
from salt.client import LocalClient
from salt.runner import RunnerClient
from multiprocessing.pool import ThreadPool


log = logging.getLogger(__name__)


def log_exception(callback, *pargs, **kwargs):
    _error_func = None
    if '_error_func' in kwargs:
        _error_func = kwargs['_error_func']
        del kwargs['_error_func']
    try:
        callback(*pargs, **kwargs)
    except Exception as e:
        log.exception("Problem deploying")
        if _error_func is not None:
            try:
                _error_func(e)
            except Exception:
                log.exception("Problem running error handler")
        raise


def deploy_background(sendmsg, tarball, project, deployment, highstate=True):
    dep = __utils__['spiro.from_pd'](project, deployment)

    sendmsg("deploy-bundle", {"msg": "Deploying bundle"})
    dep.deploy_new(tarball)

    # TODO: Do we need to flush the fileserver?

    log.debug("Highstate: %r", highstate)
    if highstate:
        runner = RunnerClient(__opts__)
        minion = LocalClient(mopts=__opts__)
        sendmsg("highstate-calc", {"msg": "Calculating minions to highstate"})
        mids = runner.cmd('spiro.query_highstate', [dep.saltenv])
        if not mids:
            sendmsg("highstate-start", {
                "msg": "No minions to highstate",
                "minions": [],
            })
            sendmsg("highstate-finish", {
                "msg": "Highstate complete",
            })
        else:
            sendmsg("highstate-start", {
                "msg": "Highstating %i minions: %s" % (len(mids), ', '.join(mids)),
                "minions": mids,
            })
            for res in minion.cmd_iter(mids, 'state.highstate', tgt_type='list', kwarg={'queue': True, 'saltenv': dep.saltenv}, expect_minions=True):
                for mid, result in res.items():
                    sendmsg("highstate-update", {
                        "msg": "Minion %s finished" % mid,
                        "minion": mid,
                        "result": result,
                    })
            sendmsg("highstate-finish", {
                "msg": "Highstate complete",
            })
    sendmsg(None, None)


@tornado.web.stream_request_body
class SpiroBuildHandler(tornado.web.RequestHandler):
    pool = ThreadPool(10)

    def prepare(self):
        super(SpiroBuildHandler, self).prepare()
        log.info("Receved Headers: %r", self.request.headers)

        token = self.get_current_user()
        if not token or not __utils__['spiro_auth.check_token'](token):
            self.set_header('WWW-Authenticate', 'Bearer')
            self.send_error(401)

    def get_current_user(self):
        try:
            auth = self.request.headers['Authorization']
        except (KeyError, TypeError):
            return
        else:
            try:
                typ, token = auth.split(' ', 1)
            except Exception:
                return
            else:
                if typ.strip().lower() != 'bearer':
                    return
                return token.strip()

    def get(self):
        self.set_header("Content-Type", "text/plain")
        self.write("Auth OK")

    @tornado.gen.coroutine
    def sse_start(self):
        """Sets up SSE for later pushing"""
        try:
            self.set_header('Content-Type', 'text/event-stream')
            self.set_header('Cache-Control', 'no-cache')
            yield self.flush()
        except tornado.iostream.StreamClosedError:
            pass

    @tornado.gen.coroutine
    def sse_push(self, data, event=None):
        """Pushes data to a listener."""
        try:
            if event is not None:
                self.write('event: %s\n' % event)
            self.write('data: %s\n\n' % json.dumps(data))
            yield self.flush()
        except tornado.iostream.StreamClosedError:
            pass

    def _check_path(self, pathname):
        # TODO: Check that this pathname is safe to extract
        return True

    @tornado.gen.coroutine
    def post(self):
        # 1. Check parameters, open tarball
        # 2. Write out files
        # 3. (Optional) Highstate touched minions
        log.debug("Received deploy")
        project = self.get_argument('project')
        deployment = self.get_argument('deployment')
        highstate = bool(self.get_argument('highstate', default=''))
        files = self.request.files['bundle']

        if not __utils__['spiro_auth.check_token'](self.get_current_user(), project):
            self.send_error(403)

        if len(files) != 1:
            self.send_error(400)  # TODO: Be more specific
        try:
            tarball = tarfile.open(fileobj=six.BytesIO(files[0].body))
            # Check validity of tarball
            bundle = tarball.getmembers()
        except tarfile.TarError:
            self.send_error(400)  # TODO: Be more specific

        if not all(self._check_path(ti.name) for ti in bundle):
            self.send_error(400)  # TODO: Be more specific

        log.debug("Starting deploy")
        # At this point, we have accepted the submission and start work
        yield self.sse_start()

        queue, sendmsg = self.mkqueue()

        def _error(exc):
            sendmsg(
                'exception',
                {'msg': ''.join(traceback.format_exception_only(type(exc), exc))}
            )
            sendmsg(None, None)

        res = self.pool.apply_async(
            log_exception,
            (deploy_background, sendmsg, tarball, project, deployment, highstate),
            {'_error_func': _error},
            # error_callback=_error,  # Py3 only
        )

        while True:
            event, data = yield queue.get()
            if event is None:
                break
            else:
                yield self.sse_push(data, event)

    def mkqueue(self):
        ioloop = tornado.ioloop.IOLoop.current()
        q = tornado.queues.Queue()

        def sendmsg(event, data):
            ioloop.add_callback(
                q.put, (event, data)
            )
        return q, sendmsg


def make_app():
    return tornado.web.Application([
        (r"/", SpiroBuildHandler),
    ])


def make_server(app, cert, key):
    if (cert and not os.path.exists(cert)) or (key and not os.path.exists(key)):
        log.error("SSL cert/key do not exist, starting without")
        cert = key = None
    if not cert:
        log.info("Clearnet server")
        return tornado.httpserver.HTTPServer(app)
    else:
        log.info("Encrypted server")
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(cert, key)

        # Protocol options: allow TLSv1.1 and later
        ssl_ctx.options |= ssl.OP_NO_SSLv2
        ssl_ctx.options |= ssl.OP_NO_SSLv3
        ssl_ctx.options |= ssl.OP_NO_TLSv1
        ssl_ctx.options |= ssl.OP_NO_TLSv1_1

        # Cipher options: strong ciphers, follow server preferences
        # From https://wiki.mozilla.org/Security/Server_Side_TLS#Modern_compatibility
        ssl_ctx.set_ciphers("ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA384:ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA256")
        ssl_ctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE

        # Key exchange: strong prime curve, no point reuse
        ssl_ctx.set_ecdh_curve('prime256v1')
        ssl_ctx.options |= ssl.OP_SINGLE_ECDH_USE

        return tornado.httpserver.HTTPServer(app, ssl_options=ssl_ctx, max_buffer_size=10*1024*1024*1024)


def start(port=4510, ssl_crt=None, ssl_key=None):
    '''
    Start the server for spiro to receive uploads from CI
    '''
    with open('/tmp/opts', 'wt') as f:
        import pprint
        pprint.pprint(__opts__, stream=f)
    log.info("Engine start: port=%r, ssl_crt=%r, ssl_key=%r", port, ssl_crt, ssl_key)
    if __opts__['__role'] == 'master':
        app = make_app()
        server = make_server(app, ssl_crt, ssl_key)
        server.listen(port)
        log.info("Starting SpiroFS Deployment Server on port %s", port)
        tornado.ioloop.IOLoop.current().start()
    else:
        # Don't do anything as the minion
        # TODO: Can we say "No, don't restart us?"
        pass



'''
Spiro fileserver backend
'''
from __future__ import absolute_import, print_function, unicode_literals
import logging

import salt.utils.gzip_util
import salt.utils.files
from salt.ext import six

log = logging.getLogger(__name__)

log.info("load")

__virtualname__ = 'spiro'


def __virtualname__():
    if __virtualname__ not in __opts__['fileserver_backend']:
        return False
    return __virtualname__


def envs():
    log.info("envs")
    return [d.saltenv for d in __utils__['spiro.list_deployments']()]


def update():
    log.info("update")


def file_list(load):
    log.info("file_list: %r", load)
    deployment = __utils__['spiro.from_env'](load['saltenv'])
    if deployment is None:
        return []
    else:
        return [
            fn
            for fn, t in deployment.contents()
            if t == 'f'
        ]


def file_list_emptydirs(load):
    log.info("file_list_emptydirs: %r", load)
    deployment = __utils__['spiro.from_env'](load['saltenv'])
    if deployment is None:
        return []
    else:
        return [
            fn
            for fn, t in deployment.contents()
            if t == 'e'
        ]


def dir_list(load):
    log.info("dir_list: %r", load)
    deployment = __utils__['spiro.from_env'](load['saltenv'])
    if deployment is None:
        return []
    else:
        return [
            fn
            for fn, t in deployment.contents()
            if t == 'd'
        ]


def symlink_list(load):
    log.info("symlink_list: %r", load)
    return {}
    # FIXME: Make mapping of link -> dest
    deployment = __utils__['spiro.from_env'](load['saltenv'])
    if deployment is None:
        return []
    else:
        return [
            fn
            for fn, t in deployment.contents()
            if t == 's'
        ]


def find_file(path, saltenv='base', **kwargs):
    log.info("find_file: %r, %r, **%r", path, saltenv, kwargs)
    dep = __utils__['spiro.from_env'](saltenv)
    if dep is None:
        return {}
    else:
        return {
            'path': path,
            'deployment': dep.serialize() if dep.exists(path) else None,
        }


def file_hash(load, fnd):
    log.info("file_hash: %r, %r", load, fnd)
    dep = __utils__['spiro.deserialize'](fnd.get('deployment'))
    if dep is None:
        return {}

    return {
        'hsum': dep.hash_file(fnd['path'], __opts__['hash_type']),
        'hash_type': __opts__['hash_type'],
    }


def serve_file(load, fnd):
    # fnd is the return of find_file()
    log.info("serve_file: %r, %r", load, fnd)
    dep = __utils__['spiro.deserialize'](fnd.get('deployment'))
    if dep is None:
        return {'data': '', 'dest': ''}

    gzip = load.get('gzip')

    with dep.open(fnd['path'], 'rb') as f:
        f.seek(load['loc'])

        data = f.read(__opts__['file_buffer_size'])
        # TBH, I'm not sure why this exists, and it causes problems
        # if data and six.PY3 and not salt.utils.files.is_binary(fnd['path']):
        #     data = data.decode(__salt_system_encoding__)
        if gzip and data:
            data = salt.utils.gzip_util.compress(data, gzip)
            return {
                'data': data,
                'dest': fnd['path'],
                'gzip': gzip,
            }
        else:
            return {
                'data': data,
                'dest': fnd['path'],
            }



import os
import pymacaroons
import random
import string
import logging
import json
import time

log = logging.getLogger(__name__)

IDENT_ALPHABET = string.digits + string.ascii_letters + string.punctuation

_key_cache = None

rand = random.SystemRandom()


def _auth_dir():
    d = os.path.join(__opts__['pki_dir'], 'spiro_auth')
    if not os.path.exists(d):
        os.makedirs(d)
    return d


def _getkey():
    global _key_cache
    if _key_cache is None:
        keyfile = os.path.join(_auth_dir(), 'macaroon-key')
        if os.path.exists(keyfile):
            with open(keyfile, 'rt') as kf:
                _key_cache = kf.read()
        else:
            _key_cache = ''.join(rand.choice(IDENT_ALPHABET) for _ in range(64))
            with open(keyfile, 'wt') as kf:
                kf.write(_key_cache)
    return _key_cache


def _genident():
    return ''.join(
        rand.choice(IDENT_ALPHABET)
        for _ in range(20)
    )


def _mkpredicate(**preds):
    return json.dumps(preds)


def _parsepredicate(txt):
    for name, arg in json.loads(txt).items():
        yield name, arg


def _check_predicates(pred, project=None, now=None):
    if now is None:
        now = time.time()
    log.debug("Check predicate (proj=%r): %r", project, pred)

    def _check_single_pred(kind, arg):
        if kind == 'not_before':
            return now > arg
        elif kind == 'not_after':
            return now < arg
        elif kind == 'project_allow':
            return project in arg or project is None
        elif kind == 'project_deny':
            return project not in arg or project is None
        else:
            log.error("Unknown predicate: %r %r", kind, arg)
            return False

    return all(
        _check_single_pred(kind, arg)
        for kind, arg in _parsepredicate(pred)
    )


def generate_token(projects=None, lifespan=None, now=None):
    m = pymacaroons.Macaroon(
        identifier=_genident(),
        key=_getkey(),
    )

    if now is None:
        now = time.time()

    caveat = {
        'not_before': now,
    }

    if lifespan is not None:
        caveat['not_after'] = now + lifespan
    if projects is not None:
        caveat['project_allow'] = projects

    m.add_first_party_caveat(_mkpredicate(**caveat))

    return m.serialize()


def check_token(token, project=None):
    """
    Checks if the given token is valid for the given project. If not project is
    given, only check that the token is valid.
    """
    try:
        m = pymacaroons.Macaroon.deserialize(token)
    except (TypeError, ValueError):  # Raised by b64decode()
        log.info("Invalid token (Invalid base64)")
        return False
    except pymacaroons.exceptions.MacaroonDeserializationException:
        log.info("Invalid token (deserialization)")
        return False

    v = pymacaroons.Verifier()
    v.satisfy_general(lambda p: _check_predicates(p, project))

    # TODO: Check m.identifier against {white,black}list

    try:
        isok = v.verify(m, _getkey())
    except pymacaroons.exceptions.MacaroonVerificationFailedException:
        log.info("Verification problem")
        # XXX: Do we want to return something more useful?
        return False
    else:
        return isok


def inspect_token(token):
    try:
        m = pymacaroons.Macaroon.deserialize(token)
    except pymacaroons.exceptions.MacaroonDeserializeException:
        return "(invalid)"

    return m.inspect()


def revoke_token(token):
    log.warning("TODO: Implement revoke_token")



from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import hashlib
import contextlib
import tempfile
import shutil
import posixpath
try:
    from urllib import unquote, quote
except ImportError:
    from urllib.parse import unquote, quote

import salt.utils.files
import salt.utils.hashutils
import salt.utils.path
import salt.utils.yamldumper
import salt.utils.yamlloader
from salt.ext import six

log = logging.getLogger(__name__)


def _fsdir(*bits):
    """
    Takes quoted arguments
    """
    # 2018.3: salt.utils.path.join()
    return os.path.join(__opts__['cachedir'], 'spirofs', *bits)


class Deployment(object):
    """
    A single environment.
    """

    def __init__(self, project, deployment):
        self.project = project
        self.deployment = deployment

    @property
    def root(self):
        """
        The root directory of this deployment.
        """
        return _fsdir(
            quote(self.project),
            quote(self.deployment),
        )

    @property
    def saltenv(self):
        """
        The saltenv for this deployment.
        """
        return "%s/%s" % (self.project, self.deployment)

    @property
    def synth_tops(self):
        """
        The name of the synthesized topfile (suitable for salt)
        """
        return __opts__['state_top']

    @property
    def real_tops(self):
        """
        The name of the spiro topfile
        """
        return 'spirotop.sls'

    @classmethod
    def from_env(cls, saltenv):
        """
        Factory: From a saltenv
        """
        if '/' not in saltenv:
            return

        parts = saltenv.rsplit('/', 1)
        self = cls(parts[0], parts[1])
        if self.exists():
            return self

    @classmethod
    def iterdeployments(cls):
        """
        Yield all Deployments
        """
        if not os.path.isdir(_fsdir()):
            return
        for p in os.listdir(_fsdir()):
            for d in os.listdir(_fsdir(p)):
                yield cls(unquote(p), unquote(d))

    def serialize(self):
        """
        Produce a JSON-suitable representation of this object
        """
        return [self.project, self.deployment]

    @classmethod
    def deserialize(cls, data):
        """
        Reverse of .serialize()
        """
        if data:
            return cls(*data)

    def exists(self, fn=None):
        """
        Check that this deployment actually exists in the filesystem.
        """
        if fn is None:
            return os.path.isdir(self.root)
        else:
            return os.path.exists(self.path(fn))

    def ensure_exists(self):
        """
        Create this deployment if it does not exist.
        """
        if not self.exists():
            os.makedirs(self.root)

    def contents(self):
        """
        Yield (name, 'f'|'d') for each filesystem entry
        """
        if not self.exists():
            return

        for path, dirs, files in salt.utils.path.os_walk(self.root):
            relpath = path[len(self.root):].strip('/')
            # FIXME: Return empty dirs
            yield relpath, 'd'
            for f in files:
                # 2018.3: salt.utils.path.join()
                yield posixpath.join(relpath, f), 'f'

    def path(self, fname):
        """
        Returns the real filesystem path for the given item.

        Doesn't actually check the filesystem.
        """
        # Kinda ignoring the fact that salt paths are always / and
        # Windows would like \ because Windows will accept /
        # 2018.3: salt.utils.path.join()
        return os.path.join(self.root, fname)

    def open(self, fname, mode='r'):
        """
        Returns a file.like object for the given itme.
        """
        return salt.utils.files.fopen(self.path(fname), mode)

    def hash_file(self, fname, hashtype):
        """
        Returns a hash for the given item.
        """
        return salt.utils.hashutils.get_hash(
            self.path(fname),
            hashtype,
        )

    def _mkbufferdir(self):
        """
        Make a temporary directory suitable for eventually making the real one
        """
        self.ensure_exists()
        return tempfile.mkdtemp(
            prefix=quote(self.deployment),
            dir=_fsdir(quote(self.project)),
        )

    def _extract(self, tarfile, dir):
        """
        Extract the given TarFile in the directory.
        """
        tarfile.extractall(dir)

    def _generate_topfile(self, dir):
        """
        Examine a staging directory and produce a topfile if a spirotop file
        exists.
        """
        # 2018.3: salt.utils.path.join()
        spirotop_fn = os.path.join(dir, self.real_tops)
        # 2018.3: salt.utils.path.join()
        top_fn = os.path.join(dir, self.synth_tops)
        if os.path.exists(top_fn):
            return
        if not os.path.exists(spirotop_fn):
            return

        with open(spirotop_fn, 'rt') as f:
            # Safe_load in 2018.3
            yaml = salt.utils.yamlloader.load(f, Loader=salt.utils.yamlloader.SaltYamlSafeLoader)

        if not yaml:
            return

        envs = {
            self.saltenv: yaml.get(self.deployment, {}),
        }

        with open(top_fn, 'wt') as f:
            salt.utils.yamldumper.safe_dump(envs, f)

    def _replace_directory(self, stagingdir):
        """
        Replace the environment with the contents of the given directory with as
        little visible intermediate state as possible.
        """
        trashdir = tempfile.mktemp(prefix=self.deployment, dir=_fsdir(quote(self.project)))
        os.rename(self.root, trashdir)
        os.rename(stagingdir, self.root)
        shutil.rmtree(trashdir)

    def deploy_new(self, tarfile):
        """
        Deploy the given TarFile as a new version of this deployment.
        """
        # 1. Set up a staging area in a place most likely to be on the same 
        #    filesystem as the serving directory
        log.debug("Make deployment buffer dir")
        stagingdir = self._mkbufferdir()

        # 2. Examine spirotop and do pre-extraction activities
        # We don't actually have anything here
        log.debug("Deployment pre-extract activities")

        # 3. Extract the tarball
        log.debug("Deployment extraction")
        self._extract(tarfile, stagingdir)

        # 4. Do post-extraction generation/compilation
        log.debug("Deployment post-extract activities")
        self._generate_topfile(stagingdir)

        # 5. Shuffle directories making the staging area the production directory
        # 6. Clean up the old version
        log.debug("Deployment directory shuffle")
        self._replace_directory(stagingdir)


def list_deployments():
    """
    Returns a list of all known Deployments (as Deployment objects).
    """
    return Deployment.iterdeployments()


def from_env(saltenv):
    """
    Gets the Deployment for the given saltenv
    """
    return Deployment.from_env(saltenv)


def from_pd(project, deployment):
    """
    Gets the Deployment for the given project/deployment pair
    """
    return Deployment(project, deployment)


def deserialize(data):
    """
    Turns the results of Deployment.serialize() back into an object.
    """
    return Deployment.deserialize(data)



import os.path

tags = (
    'auth', 'beacons', 'cache', 'engines', 'executor', 'fileserver', 'grains',
    'log_handlers', 'module', 'pillar', 'render', 'returner', 'runner',
    'serializers', 'tokens', 'top', 'utils', 'wheel',
)


class Loader:
    def __init__(self, dirname):
        self.dirname = dirname
        self.root = os.path.dirname(os.path.abspath(__file__))

    def __call__(self):
        fullpath = os.path.join(self.root, self.dirname)
        if os.path.isdir(fullpath):
            yield fullpath


for tag in tags:
    globals()[tag] = Loader(tag)
